package ch.admin.bit.avam.officeintegration.business.xml;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

import javax.xml.bind.JAXBException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.docx4j.XmlUtils;
import org.docx4j.customXmlProperties.DatastoreItem;
import org.docx4j.jaxb.Context;
import org.docx4j.model.datastorage.BindingHandler;
import org.docx4j.model.datastorage.CustomXmlDataStorage;
import org.docx4j.model.datastorage.CustomXmlDataStorageImpl;
import org.docx4j.openpackaging.exceptions.Docx4JException;
import org.docx4j.openpackaging.exceptions.InvalidFormatException;
import org.docx4j.openpackaging.io.LoadFromZipNG;
import org.docx4j.openpackaging.io.SaveToZipFile;
import org.docx4j.openpackaging.packages.WordprocessingMLPackage;
import org.docx4j.openpackaging.parts.CustomXmlDataStoragePart;
import org.docx4j.openpackaging.parts.CustomXmlDataStoragePropertiesPart;
import org.docx4j.openpackaging.parts.WordprocessingML.MainDocumentPart;
import org.docx4j.wml.SdtPr;
import org.docx4j.wml.Tag;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Utility class for docx formats
 * 
 * @author Adrian Fortuzi
 */
public class OfficeXmlUtils {

	 
	public ByteArrayOutputStream applyArray(String [] valueArray, String [] fieldNames, InputStream inputStream) throws Exception {
		ByteArrayOutputStream oStream = new ByteArrayOutputStream();       
		
		Document customDoc = null;
		WordprocessingMLPackage wordMLPackage = null;
		wordMLPackage = getWordMLPackage(inputStream);

		// Create new customXml part, defaults to /customXml/item1.xml
		// If for example /customXml/item1.xml already exists, this allows the name /customXml/item2.xml to be generated.
		CustomXmlDataStoragePart customXmlDataStoragePart = new CustomXmlDataStoragePart(wordMLPackage.getParts());

		// The actual XML data
		CustomXmlDataStorage data = new CustomXmlDataStorageImpl();		
		customDoc = createCustomXmlDocument(valueArray, fieldNames);
		data.setDocument(customDoc);		
		customXmlDataStoragePart.setData(data);
		

		wordMLPackage.getMainDocumentPart().addTargetPart(customXmlDataStoragePart);
		String dataBindingStoreItemId = addProperties(customXmlDataStoragePart);
		wordMLPackage.getCustomXmlDataStorageParts().put(dataBindingStoreItemId.toLowerCase(), customXmlDataStoragePart ); 
		
		// Now save it 		
//		String saveFile00 = "C:\\Projects\\AVAM\\officeXML\\CC_Created_saved.docx";
//		File newFile00 = new File(saveFile00);
//		wordMLPackage.save(newFile00);
		
		// Go through all the sdt's
		MainDocumentPart documentPart = wordMLPackage.getMainDocumentPart();		
		String xpathSdt = "//w:sdt";
		List<Object> list = null;
		try {
			list = documentPart.getJAXBNodesViaXPath(xpathSdt, false);
		} catch (JAXBException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		

		int i = 0;
		for(Object raw : list) {
			i++;
			//System.out.println("raw.getClass().getName(): " + raw.getClass().getName() );			
			Object o = XmlUtils.unwrap(raw);
			Method m = o.getClass().getDeclaredMethod("getSdtPr");			
			SdtPr sdtPr = (SdtPr)m.invoke(o);			
			Tag tag = sdtPr.getTag();
										
			if (sdtPr.getDataBinding()== null) {									
				if (tag==null) continue;
				String xpath = "/root/" + tag.getVal();		
				org.docx4j.wml.CTDataBinding dataBinding = Context.getWmlObjectFactory().createCTDataBinding();				
				dataBinding.setStoreItemID(dataBindingStoreItemId);
				dataBinding.setXpath(xpath);
				dataBinding.setParent(sdtPr);
				//dataBinding.setPrefixMappings("n0");				
				sdtPr.setDataBinding(dataBinding);					
			}			
		}

//		CustomXmlDataStoragePart part = wordMLPackage.getCustomXmlDataStorageParts().get(dataBindingStoreItemId.toLowerCase());			
//		if (part==null) {
//			System.out.println("Couldn't locate part by storeItemId " + dataBindingStoreItemId);
//		}
									
		// Apply the bindings
		BindingHandler.applyBindings(wordMLPackage.getMainDocumentPart());
		//System.out.println(XmlUtils.marshaltoString(wordMLPackage.getMainDocumentPart().getJaxbElement(), true, true));
		
		//? Now save it 		
//		String saveFile = "C:\\Projects\\AVAM\\officeXML\\CC_Created.docx";
//		wordMLPackage.save(new java.io.File(saveFile));
		
		SaveToZipFile saver = new SaveToZipFile(wordMLPackage);
        saver.save(oStream);
        
		return oStream;
	}
	
	
	public String[] readContentControls(InputStream docAsStream) {

		//String fields[] = null;
		ArrayList<String> fields = new ArrayList<String>();
		
		WordprocessingMLPackage wordMLPackage = null;
		wordMLPackage = getWordMLPackage(docAsStream);
			    
		//HashMap<String, String> contentControls = new HashMap<String, String>();

		MainDocumentPart mainPart = wordMLPackage.getMainDocumentPart();

		String xpathSdt = "//w:sdt/w:sdtPr/w:tag";
		List<Object> list = null;
		try {
			list = mainPart.getJAXBNodesViaXPath(xpathSdt, false);
		} catch (JAXBException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		for (Iterator<Object> it = list.iterator(); it.hasNext();) {
			Object o = XmlUtils.unwrap(it.next());
			if (o instanceof Tag) {
				Tag tag = (Tag) o;
				String val = tag.getVal();				
				fields.add(val);				
				//System.out.println("Tag.Val: " + val);
			}
		}
		
		String[] strArray = new String[fields.size()];
		return fields.toArray(strArray);

	}

	private WordprocessingMLPackage getWordMLPackage (InputStream docAsStream) {
		
		LoadFromZipNG fileLoader = new LoadFromZipNG();
	    //OpcPackage content = loader.get(docAsStream);
	    
		WordprocessingMLPackage wordMLPackage = null;
		try {
			wordMLPackage = (WordprocessingMLPackage) fileLoader.get(docAsStream);
		} catch (Docx4JException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	
		return wordMLPackage;
	}

	
	private org.w3c.dom.Document createCustomXmlDocument(String [] valueArray, String [] fieldNames) {

		Document doc = null;
		try {
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();

			// root elements
			doc = docBuilder.newDocument();
			Element rootElement = doc.createElement("root");
			doc.appendChild(rootElement);

			Hashtable<String, String> contentControls = new Hashtable<String, String>();
			String valueContentControl = "";
			for (int count = 0; count < fieldNames.length; count++) {
				//System.out.println("fieldNames[" + count + "] = " + fieldNames[count]);
				//System.out.println("valueArray[" + count + "] = " + valueArray[count]);
				// CC elements
				if (valueArray[count] ==  null || valueArray[count].isEmpty() || valueArray[count].length() < 1) {
					valueContentControl = "";
				} else {
					valueContentControl = valueArray[count];
				}								
				contentControls.put(fieldNames[count], valueContentControl);			
			}
			
			Enumeration<String> contentControlKey = contentControls.keys();
			while(contentControlKey.hasMoreElements()) {
				String key = contentControlKey.nextElement();
				Element contentControlElement = doc.createElement(key);
				contentControlElement.appendChild(doc.createTextNode(contentControls.get(key)));
				rootElement.appendChild(contentControlElement);
				System.out.println("key = " + key + "; value = " + contentControls.get(key));
			}

			/*
			// write the content into xml file
			String outputXmlPath = "C:\\Projects\\AVAM\\officeXML\\Vorlagen\\ZG_files\\CustomXml.xml";			                       			
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File(outputXmlPath));
			transformer.transform(source, result);
*/
			System.out.println("Done - CustomXml.xml ");

		} catch (ParserConfigurationException pce) {
			pce.printStackTrace();
	//	} catch (TransformerException tfe) {
		//	tfe.printStackTrace();
		}

		return doc;

	}

	// 
	public String addProperties(CustomXmlDataStoragePart customXmlDataStoragePart) throws InvalidFormatException {

		CustomXmlDataStoragePropertiesPart part = new CustomXmlDataStoragePropertiesPart();
		
		org.docx4j.customXmlProperties.ObjectFactory of = new org.docx4j.customXmlProperties.ObjectFactory();

		DatastoreItem dsi = of.createDatastoreItem();
		String newItemId = "{" + UUID.randomUUID().toString() + "}";
		dsi.setItemID(newItemId);
		part.setJaxbElement(dsi);
		customXmlDataStoragePart.addTargetPart(part);
		
		return newItemId;
	}
	
}
